/*
Class:   GridCombobox
Author:  Daniel Howard (sproket@total.net)
History: 
   1.0  Mar 27, 1999 
      Initial release

   1.1  Mar 28, 1999
      Added handler for enter key in entryfield so the user can type
      some keys to jump to a record and press enter to select it.

   1.1a Mar 28, 1999
      Removed blue from grid and futzed a bit with the enter key
      behavior.

   1.1b Mar 29, 1999
      Added form.Repaint() with each togglegrid call       

   1.2 July 1, 2000
      *  Changed metric to Pixel
      *  Added onChange event handler
      *  Added RefreshAlways property to refresh controls
      *  Added SetValue method to centralize the setting of the value 
         of the entryfield
      *  Renamed base form to DHForm so as not to conflict with a 
         possible dFORM from dBASE Inc.
      *  Moved form.Repaint() to the ToggleGrid method
      *  Fixed tabbing order problem.

Purpose:
   Use this object like a combobox except the dropdown is not a 
   simple list but a grid object. 

   This class requires:
   DHForm.cfm      - Base form class 
   down.bmp        - Down Arrow bitmap 


   Properties:

      DropDownHeight -  Specifies the dropdown height of the Grid in
      					   PIXEL metrics.

		IsGridOpen -		Tells you whether the Grid is currently open or
                        not.

      CarryFontInformation - 
                        If set to true (the default), the font 
                        information you specify in the entryfield will
                        be carried over into the grid columns.

      RefreshAlways -   If set to True, calls form.rowset.refreshControls() 
                        when the value changes. (default is true)


Usage:

   Drop one on your form and datalink the grid to any rowset object. 
   You must also define at least one Column object or you'll get an 
   error.
   

Notes:
   - This doesn't behave exacly like a regular combobox. Improvement
     in this area would be greatly appreciated.

   - Currently this only works with metric PIXELS.

   - The other classes in this file are used by this control you 
     should not use them as stand-alone classes.

   - the gd_seeker was _stolen_ from the Borland samples. I made
     a few mods to it though...

   - Wouldn't it be nice to have a NOPALET keyword on the class
     that would prevent these child classes from appearing.          

   - Doesn't yet handle LookupSQL type of combobox.

   - Creates an _app.GridCombo as a pointer to "this". Needed for
     the arrow keys.  It gets nullified when the entryfield loses
     focus.

This software is donated to the public domain. You may modify 
this code in anyway you wish however the author reserves the 
right to whatever official version is published. If you have 
further ideas/bug reports/additions then contact the author and 
your name will be added to the contributors if your idea is 
implemented.

The author accepts no responsibility for the use or consequences 
of the use of this software.
*/

class GridCombobox(parentObj) of CONTAINER(parentObj) custom
   with (this)
      left = 49
      top = 77
      width = 301
      height = 31
      metric = 6	// Pixels
      borderStyle = 4	// Single
   endwith

   this.ENTRYFIELD = new GC_Seeker(this)
   with (this.ENTRYFIELD)
      height = 20
      left = 0
      top = 0
      width = 298
      metric = 6	// Pixels
      fontSize = 9
      value = ""
   endwith

   this.GRID = new GC_GRID(this)
   with (this.GRID)
      bgColor = "white"
      cellHeight = 20
      pageno = 0
      height = 8
      left = 0
      top = 21
      width = 298
      metric = 6	// Pixels
   endwith


   this.IsGridOpen = false
   this.onChange = NULL
   this.DropDownHeight = 220
   this.CarryFontInformation = true
   this.RefreshAlways = true


   Function Init
      // Hide the container's border at run-time.
      this.BorderStyle = 3 // None

      // Reposition the entryfield to the size of the container
      this.EntryField.Width = this.Width - 3

      // Assign the rowset to the seeker
      if type("this.Grid.DataLink") # "U"
         // rowset to perform seeker operations
         this.EntryField.rowset := this.Grid.DataLink
      endif

      // Create the drop button
      this.DropBUTTON = new PUSHBUTTON(this)
      with (this.DropButton)
         onClick = class::DropButton_ONCLICK
         borderStyle = 2
         metric = 6
         height = this.EntryField.height - 2
         width = 20
         left = this.EntryField.Width - width - 1
         top = this.EntryField.Top + 1
         text = ""
         speedBar = true
         upBitmap = "filename Down.bmp"
         fontSize = 9
         pageno = 0
         value = false
      endwith

      // Prototype the WinAPI I'll need
      if type('WinAPISetParent') # 'FP'
         extern chandle WinAPISetParent(chandle, chandle) user32 from 'SetParent'
      endif

      // Fix top & Left
      this.Grid.Top += this.Top 
      this.Grid.Left += this.Left 
      this.Grid.Width := this.EntryField.Width
      this.Grid.Height := this.DropDownHeight

      // If this property is true then carry all of the font
      // information from the entryfield and assign it to the
      // grid columns.
      if this.CarryFontInformation
         // set the cellheight to the entryfield height less a bit.
         this.Grid.CellHeight = this.EntryField.Height - 1

         local f,j
         set procedure to font.cc additive
         f = New Font()
         f.DefineFrom(this.EntryField)
         for j = 1 to this.Grid.Columns.Size
            f.SetObjectFont(this.Grid.Columns["Column"+j].EditorControl)
         next j
      endif
   return


   function DropButton_onClick
      this.Parent.ToggleGrid()
   return

   Function ToggleGrid
      if not this.IsGridOpen
         this.Grid.Visible = false // make sure I'm hidden
         local bm
         bm = this.Grid.DataLink.BookMark()
         // Kill any OnSelChange Event defined because it will
         // fire here. We don't want that since the user of this
         // component will probably use onSelChange for his/her
         // own purposes. This blop of code is really just to 
         // scroll the grid properly.
         local oldOnSelChange
         oldOnSelChange = this.Grid.OnSelChange
         this.Grid.OnSelChange := NULL
         this.Grid.DataLink.First()
         this.Grid.DataLink.Goto(bm)
         this.Grid.OnSelChange := oldOnSelChange
         this.Grid.Visible := true
         WinApiSetParent(this.Grid.hWnd,form.hWnd)
         this.EntryField.SetFocus()
      else
         this.Grid.Visible := false
      endif

      this.IsGridOpen := not this.IsGridOpen
      form.Repaint()
   return

   function MoveRecord(n)
      this.grid.Datalink.next(n)
      // updates the entryfield....
      this.SetValue()
   return


   function SetValue
      this.EntryField.Value := this.grid.Columns["Column1"].DataLink.Value

      if this.RefreshAlways
         if type("form.rowset") # "U"
            form.rowset.refreshControls()
         endif
      endif

      if type("this.onChange") # "U"
         this.onChange( (this) )
      endif
   return

endclass


CLASS GC_Seeker( oCont ) of Entryfield( oCont ) custom
protect lNear, lExact, lFound, lRefresh, nMaxPos, onLeftMouseUp

   this.FontSize := 9

   this.onLeftMouseUp = Class::OnLeftMouseUp

   // Set stock properties
   this.selectAll := true
   this.value     := ""
  
   // Custom properties
   this.rowset    = null    // The seeker's rowset
   this.tabAfter  = 0       // Automatically tab after typing N chars
   this.seek      = this.oSeek

   // Internal properties
   this.lNear     = ( SET( "NEAR"  ) == "ON" )
   this.lExact    = ( SET( "EXACT" ) == "ON" )
   this.lRefresh  = this.form.RefreshAlways
   this.lFound    = true
   this.nMaxPos   = 0       // Max position of cursor in entryfield

   FUNCTION onGotFocus
      this.seek     = this.oSeek
      this.seekRowset = this.rowset
      this.lFound  = .t.
      this.nMaxPos = 0

      _app.GridCombo = this.parent
      // arrow handler
      on key label Uparrow _app.GridCombo.MoveRecord(-1)
      on key label Dnarrow _app.GridCombo.MoveRecord(+1)
   return

  FUNCTION key( nChar, nPosition )

    if nChar = 9
      // handle TABbing out
      if this.Parent.IsGridOpen
         this.parent.SetValue()
         this.Parent.ToggleGrid()   // Close

         // check to see if we can move the cursor to the next object...
         if type("this.Parent.before.SetFocus") # "U"
            this.Parent.before.SetFocus()
         elseif type("this.Parent.before.first.SetFocus") # "U"
            this.Parent.before.first.SetFocus()
         endif            
         return nChar
      else
         return nChar
      endif                  
    endif


    if not this.Parent.IsGridOpen and nChar # 13
      this.Parent.ToggleGrid()   // Open it
    endif


    if nextkey() # 0
      // Wait until there are no keystrokes pending, and...
      RETURN ( nChar # 255 )     // ...ignore keystroke if special key
    endif
    if nChar == 255
      if .not. empty( this.value ) .and. ;
          ( this.lFound .or. nPosition <= this.nMaxPos )
        this.lFound = this.seek()
      endif
      if this.tabAfter > 0
        // Stuff tab if position is greater than tabAfter
        nChar = iif( nPosition > this.tabAfter, 9, 0 )
      else
        nChar = 0
      endif
      this.nMaxPos = max( nPosition, ;
          min( this.nMaxPos, len( trim( this.Value ))) )
    else
      if nChar >= 32 and nChar < 255
        // Stuff special key to refire key event
        keyboard chr(255)
      elseif nChar = 13 
         this.parent.SetValue()
         this.Parent.ToggleGrid()
      endif

    endif
 RETURN nChar


  FUNCTION oSeek
    local lRet
    lRet = this.rowset.findKeyNearest( this.normalizedValue() )
    if ( this.rowset.endOfSet )
      this.rowset.next( -1 )
    endif
    RETURN lRet

  FUNCTION normalizedValue
    RETURN upper( trim( this.value ))

   FUNCTION onLostFocus    
      if form.rowset == null
         if .not. this.lNear
            set near off
         endif
         if this.lExact
            set exact on
         endif
         form.RefreshAlways = this.lRefresh
      endif

      // arrow handler
      _app.GridCombo = null
      on key label Uparrow 
      on key label Dnarrow 
   return


   Function onLeftMouseUp
      this.Parent.ToggleGrid()
   return

ENDCLASS



Class GC_Grid(oParent) of Grid(oParent)
   with (this)
      metric = 6	// Pixels
      gridLineWidth = 0
      hasColumnLines = false
      hasRowLines = false
      hasIndicator = false
      hasColumnHeadings = false
      hasVScrollHintText = false
      rowSelect = true
      allowRowSizing = false
      allowColumnSizing = false
      allowColumnMoving = false
      allowEditing = false
      allowAddRows = false
      hScrollBar = 0	// Off
      visible = false
   endwith

   function onLeftMouseUp
      this.parent.SetValue()
      this.parent.ToggleGrid()
   return


endclass



